package org.blue.club.services;

import com.mybatisflex.core.paginate.Page;
import jakarta.servlet.ServletOutputStream;
import jakarta.servlet.http.HttpServletResponse;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.tomcat.util.buf.HexUtils;
import org.blue.club.dao.UserDao;
import org.blue.club.dao.redis.AvatarOperationDao;
import org.blue.club.dao.redis.VerifyDao;
import org.blue.club.entities.dto.User;
import org.blue.club.entities.dto.req.*;
import org.blue.club.entities.dto.resp.VerifyCodeResponse;
import org.blue.club.entities.vo.data.AvatarTmpVo;
import org.blue.club.entities.vo.data.UserVo;
import org.blue.club.entities.vo.VerifyVo;
import org.blue.club.utils.FileUtils;
import org.blue.club.utils.VerifyCodeUtils;
import org.mmga.spring.boot.starter.entities.PagerData;
import org.mmga.spring.boot.starter.entities.Result;
import org.mmga.spring.boot.starter.exception.AuthorizationException;
import org.mmga.spring.boot.starter.utils.RandomUtils;
import org.mmga.spring.boot.starter.utils.VoUtils;
import org.springframework.http.ContentDisposition;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;
import org.springframework.web.multipart.MultipartFile;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.util.Date;
import java.util.List;
import java.util.Optional;

import static com.mybatisflex.core.query.QueryMethods.select;
import static org.blue.club.entities.vo.data.table.UserVoTableDef.USER_VO;

@Slf4j
@Service
@RequiredArgsConstructor
public class UserServices {
    private final UserDao userDao;
    private final VerifyDao verifyDao;
    private final VoUtils voUtils;
    private final RandomUtils randomUtils;
    private final File tmpFolder;
    private final File avatarFolder;
    private final AvatarOperationDao avatarOperationDao;
    private final FileUtils fileUtils;
    private final VerifyCodeUtils verifyCodeUtils;

    private boolean isWrongVerifyCode(String key, String code) {
        Optional<VerifyVo> targetCode = verifyDao.findById(key);
        if (targetCode.isEmpty()) {
            return true;
        }
        verifyDao.deleteById(key);
        return !targetCode.get().answer().equals(code);
    }

    public Result<User> login(LoginRequest loginRequest) {
        if (this.isWrongVerifyCode(loginRequest.key(), loginRequest.code())) {
            return Result.failed(HttpStatus.FORBIDDEN, "错误的验证码");
        }
        UserVo userVo = userDao.selectOneWithRelationsByCondition(USER_VO.NAME.eq(loginRequest.username()).and(USER_VO.PASSWORD.eq(loginRequest.sha1HexPassword())));
        if (userVo == null) return Result.failed(HttpStatus.NOT_FOUND, "用户不存在！");
        User user = voUtils.vo2DtoSafe(userVo, User.class);
        return Result.success(user);
    }

    public Result<User> createUser(RegisterRequest request) {
        if (this.isWrongVerifyCode(request.key(), request.code())) {
            return Result.failed(HttpStatus.FORBIDDEN, "错误的验证码");
        }
        if (userDao.selectCountByCondition(USER_VO.NAME.eq(request.username())) > 0) {
            return Result.failed(HttpStatus.CONFLICT, "用户已存在");
        }
        UserVo userVo = new UserVo();
        userVo.setName(request.username());
        userVo.setPassword(request.sha1HexPassword());
        userVo.setAuth(request.auth());
        userDao.insert(userVo);
        UserVo insertedUserData = userDao.selectOneWithRelationsById(userVo.getId());
        return Result.success(voUtils.vo2DtoSafe(insertedUserData, User.class));
    }

    public Result<PagerData<User>> getAllUserInfo(int num, int page) {
        Page<UserVo> userVoPage = userDao.paginateWithRelations(page, num, select(USER_VO.ALL_COLUMNS).from(USER_VO));
        long totalRow = userVoPage.getTotalRow();
        List<UserVo> records = userVoPage.getRecords();
        return Result.success(new PagerData<>(totalRow, records.stream().map(e -> voUtils.vo2DtoSafe(e, User.class)).toList()));
    }

    public Result<Boolean> changeUsername(UserRenameRequest renameVo, User user) {
        Long id = renameVo.getId();
        Long operatorId = user.getId();
        if (id == null) id = operatorId;
        if (!operatorId.equals(id) && !user.hasPermission(4L))
            throw new AuthorizationException(Result.failed(HttpStatus.UNAUTHORIZED, "权限不足！"));
        String s = renameVo.getNewName();
        if (userDao.selectCountByCondition(USER_VO.NAME.eq(s)) > 0)
            return Result.failed(HttpStatus.CONFLICT, "用户名已存在！");
        UserVo userVo = new UserVo();
        userVo.setId(id);
        userVo.setName(s);
        userDao.update(userVo);
        return Result.success(true);
    }

    public Result<Boolean> changePassword(UserChangePasswordRequest changePasswordVo, User user) {
        UserChangePasswordRequest shaHex = changePasswordVo.sha1HexPassword();
        String newPassword = shaHex.getNewPassword();
        String oldPassword = shaHex.getOldPassword();
        Long id = changePasswordVo.getId();
        UserVo newUserVo = new UserVo();
        newUserVo.setId(id);
        newUserVo.setPassword(newPassword);
        if (user.hasPermission(4L)) {
            userDao.update(newUserVo);
            return Result.success(true);
        }
        Long operatorUserId = user.getId();
        if (oldPassword == null || !id.equals(operatorUserId))
            throw new AuthorizationException(Result.failed(HttpStatus.UNAUTHORIZED, "权限不足！"));
        if (userDao.selectCountByCondition(USER_VO.ID.eq(operatorUserId)) <= 0)
            return Result.failed(HttpStatus.NOT_FOUND, "旧密码不匹配");
        userDao.update(newUserVo);
        return Result.success(true);
    }

    public Result<Boolean> changeAuth(UserChangeAuthRequest userChangeAuthVo) {
        UserVo userVo = new UserVo();
        userVo.setId(userChangeAuthVo.id());
        userVo.setAuth(userChangeAuthVo.authId());
        userDao.update(userVo);
        return Result.success(true);
    }

    @SneakyThrows
    public Result<String> changeAvatar(MultipartFile file) {
        String s = randomUtils.generatorRandomFileName(32);
        DigestInputStream digestInputStream = new DigestInputStream(file.getInputStream(), DigestUtils.getSha1Digest());
        File tmpFile = new File(tmpFolder, s);
        try (FileOutputStream fos = new FileOutputStream(tmpFile)) {
            digestInputStream.transferTo(fos);
        }
        MessageDigest messageDigest = digestInputStream.getMessageDigest();
        String sha1Hex = HexUtils.toHexString(messageDigest.digest());
        avatarOperationDao.save(new AvatarTmpVo(s, sha1Hex, new Date()));
        return Result.success(s);
    }

    public Result<Boolean> changeAvatar(String avatarOperationCode, User user) {
        Optional<AvatarTmpVo> byId = avatarOperationDao.findById(avatarOperationCode);
        if (byId.isEmpty()) return Result.failed(HttpStatus.NOT_FOUND, "操作码错误或过期");
        AvatarTmpVo avatarTmpVo = byId.get();
        File tmpFile = new File(tmpFolder, avatarOperationCode);
        String sha1 = avatarTmpVo.sha1();
        File targetFile = new File(avatarFolder, sha1);
        Optional<Result<Boolean>> booleanResult = changeAvatarFile(tmpFile, targetFile);
        if (booleanResult.isPresent()) return booleanResult.get();
        avatarOperationDao.deleteById(avatarOperationCode);
        UserVo userVo = new UserVo();
        userVo.setId(user.getId());
        userVo.setAvatar(user.getAvatar());
        userDao.update(userVo);
        return Result.success(true);
    }

    public Optional<Result<Boolean>> changeAvatarFile(File tmpFile, File targetFile) {
        if (targetFile.exists()) {
            fileUtils.tryDeleteOrDeleteOnExit(targetFile);
            return Optional.empty();
        }
        if (!tmpFile.renameTo(targetFile))
            return Optional.of(Result.failed(HttpStatus.INTERNAL_SERVER_ERROR, "移动头像文件失败！请联系管理员处理"));
        return Optional.empty();
    }

    public Result<VerifyCodeResponse> getVerifyCode() {
        VerifyCodeResponse verifyCodeResponse = verifyCodeUtils.generateVerifyCode();
        return Result.success(verifyCodeResponse);
    }

    @SneakyThrows
    public void getAvatar(HttpServletResponse response, String sha1) {
        File targetFile = new File(avatarFolder, sha1);
        if (!targetFile.getAbsolutePath().startsWith(avatarFolder.getAbsolutePath())) {
            Result.failed(HttpStatus.NOT_FOUND, "文件不存在！").writeToResponse(response);
            return;
        }
        if (!targetFile.exists()) {
            Result.failed(HttpStatus.NOT_FOUND, "文件不存在！").writeToResponse(response);
            return;
        }
        try (FileInputStream fileInputStream = new FileInputStream(targetFile)) {
            response.setContentType("image/png");
            response.setHeader("Content-Disposition", ContentDisposition.attachment().filename(sha1 + ".png").build().toString());
            ServletOutputStream outputStream = response.getOutputStream();
            StreamUtils.copy(fileInputStream, outputStream);
        }
    }

    public Result<User> getUserInfo(int userId) {
        UserVo userVo = userDao.selectOneWithRelationsById(userId);
        if (userVo == null) return Result.failed("用户不存在！");
        User user = voUtils.vo2DtoSafe(userVo, User.class);
        return Result.success(user);
    }
}
